@@ -1,10 +1,11 @@
 __all__ = ['thread',
            'thread_pool',
            'process',
+           'ThreadPool',
            'PebbleError',
            'TimeoutError',
            'TaskCancelled']
 
 from .pebble import PebbleError, TimeoutError
 from .process import process, TaskCancelled
-from .thread import thread, thread_pool
+from .thread import thread, thread_pool, ThreadPool
